//We always have to include the library

// OSC Library section
#include <OSCBoards.h>
#include <OSCMessage.h>
#include <SLIPEncodedSerial.h>
#include <SLIPEncodedUSBSerial.h>

#ifdef BOARD_HAS_USB_SERIAL
#include <SLIPEncodedUSBSerial.h>
SLIPEncodedUSBSerial SLIPSerial( thisBoardsSerialUSB );
#else
#include <SLIPEncodedSerial.h>
SLIPEncodedSerial SLIPSerial(Serial);
#endif
// OSC Library section end


#define sensorNum 5
// Analog sensor value pressure
int analogVal[sensorNum] = {
  0, 0, 0, 0, 0
};
// Analog sensor pins pressure
const char A_pins[sensorNum] = {
  A0, A1, A2, A3, 9
};

int16_t sensor_sum = 0;

// time guard pressure sensors
unsigned long fps_ts;
const int fps_interval = 50;

// RFID global
#include"rf_packets.h"
// Use global for now
packet_rcv packetRCV;
uint8_t data_to_read = 0;
uint8_t data_ptr = 0;
int16_t rcv_Packet_Stat = -1;

int32_t rcv_Mifare1K_ID;
// RFID golbal end

// Debug
// #define DEBUG_RF

// Distance sensor
const char Dist_pins[] = {A6, A7};
unsigned long ir_dist_ts;
const int ir_dist_interval = 25;
// Distance sensor end

// Blow piezo
//const int blowPin = 9;
//unsigned long ts_startblow;
//const int blow_length = 200;
//
//boolean has_blow = false;
//// Countdown
//int samples_got = 0;
//const int sample_trg = 200;
// Blow end


// send packet protection
boolean packet_otw = false;
unsigned long packet_ts;
const int packet_interval = 10;
// prot end

// Neopixel
#include <Adafruit_NeoPixel.h>
#define PIN            8
#define NUMPIXELS      8
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

boolean neo_ID = false;

char neo_stat = -1;
char neo_pos = 0;

const float decay_ratio = 0.95f;
const int decay_thresh = 10;

unsigned long neo_ts;
const int neo_interval = 80;

unsigned long neo_dimm_ts;
const int neo_dimm_interval = 20;
// Neopixels end

void setup() {
  randomSeed( analogRead(A6) + analogRead(A7) );

  // pixels
  pixels.begin();
  pixels.clear();

  Serial1.begin(57600); // RFID Baud rate

  // Init distance sensor
  for (int i = 0; i < 2; i++ ) {
    pinMode( Dist_pins[i], INPUT );
  }

  // Debug
#ifdef DEBUG_RF
  Serial.begin(115200); // Use serial as debug
  while ( !Serial ); // Leonardo Serial bug
#endif

  // Use serial as OSC SLIP
  SLIPSerial.begin(115200);

  while ( !Serial ); // Leonardo Serial bug


  // Selftest
  for ( int i = 0; i < 10; i++ ) {
    for ( int j = 0; j < NUMPIXELS; j++ ) {
      pixels.setPixelColor(j, pixels.Color(random(64), random(64), random(64)));
    }
    pixels.show();
    delay(100);
    pixels.clear();
    pixels.show();
    delay(100);
  }
}


void loop() {
  if ( int size = Serial1.available() > 0 ) { // there's data
    while ( size-- ) {
      form_Packet();
    }
  }

  if ( packet_finished() ) { // complete packet
    // process packet here;

    packet_finished_clear();
  }

  if ( mf1k_id() ) {
    // process Mifare 1k id here
#ifdef DEBUG_RF
    Serial.print( "Mifare 1K detected, ID : 0x" );
    Serial.println(rcv_Mifare1K_ID, HEX);
#endif
    if ( !packet_is_sending() ) { // no packet is sending
      packet_sending(); // lock packet

      // ID sensed but not consumed
      neo_ID = true;

      OSCMessage msg("/ID");
      msg.add((int32_t)rcv_Mifare1K_ID);

      SLIPSerial.beginPacket(); // mark the beginning of the OSC Packet
      msg.send(SLIPSerial);
      SLIPSerial.endPacket();

      packet_finished_clear();
    }
  }

  // Blow
  //  process_blow();

  // Sense pressure (Independent from RFID)
  process_pressure();

  // Distance sensor
  process_dist();

  // LED
  process_neopixel();
}

// MF1K packet
void form_Packet() { // Hard coded for now
  uint8_t read_val;
  uint8_t read_chksum;
  switch (rcv_Packet_Stat) {
    case -1:
      read_val = Serial1.read();
      if ( read_val == 0x04 ) { // Mifare 1k
        rcv_Packet_Stat = 10; // this means it's the reply of auto poll
        data_to_read = 4;
        rcv_Mifare1K_ID = 0x00000000; // clear ID
#ifdef DEBUG_RF
        Serial.println( "Not reply, auto poll, Mifare 1K detected" );
#endif
        break;
      }
      if ( read_val == 0x55 ) {
        rcv_Packet_Stat = 0;
        packetRCV.chksum = 0x55;
      }
      break;
    case 0:
      if ( Serial1.read() == 0x55 ) {  // Packet really started
        packetRCV.length = 0x0000;  // Clear
        rcv_Packet_Stat = 1;
        packetRCV.chksum ^= 0x55;
      } else {
        rcv_Packet_Stat = -1;
      }
      break;
    case 1: // Read reserve
      packetRCV.reserve = Serial1.read();
#ifdef DEBUG_RF
      Serial.print("Packet Head - reserve byte 0x");
      Serial.println(packetRCV.reserve, HEX);
#endif
      packetRCV.chksum ^= packetRCV.reserve;
      rcv_Packet_Stat = 2;
      break;
    case 2: // Read address
      packetRCV.address = Serial1.read();
#ifdef DEBUG_RF
      Serial.print("Packet Address - ");
      Serial.println(packetRCV.address, HEX);
#endif
      packetRCV.chksum ^= packetRCV.address;
      rcv_Packet_Stat = 3;
    case 3: // Read length
    case 4:
      read_val = Serial1.read();
      packetRCV.length |= read_val << (rcv_Packet_Stat == 3) ? 8 : 0;
      packetRCV.chksum ^= read_val;
      data_to_read = packetRCV.length - 1;
      data_ptr = 0;
      rcv_Packet_Stat++;
#ifdef DEBUG_RF
      if ( rcv_Packet_Stat == 4 ) {
        Serial.print("Req Length - ");
        Serial.println( packetRCV.length );
      }
#endif
      break;
    case 5: // read stat
      packetRCV.result = Serial1.read();
      packetRCV.chksum ^=  packetRCV.result;
      data_to_read--;
#ifdef DEBUG_RF
      Serial.print("Reply Stat - ");
      Serial.println( (packetRCV.result == RF_SUCCESS) ? "Succeed" : "Failed" );
#endif
      rcv_Packet_Stat = 6;
      break;
    case 6: // read data
      if ( data_to_read <= 0 ) { // no data left, continue to next
        rcv_Packet_Stat = 7;
        break;
      }
      packetRCV.data[data_ptr] = Serial1.read();
      packetRCV.chksum ^= packetRCV.data[data_ptr];
      data_ptr++;
      data_to_read--;
      break;
    case 7: // read chksum
      read_chksum = Serial1.read();
      rcv_Packet_Stat = 8;
#ifdef DEBUG_RF
      Serial.print( "Checksum :" );
      Serial.print( packetRCV.chksum, HEX );
      Serial.print( "| Received :" );
      Serial.print( read_chksum );
      Serial.println( (read_chksum == packetRCV.chksum) ? "Same" : "Error" );
#endif;
      break;
    case 8:
#ifdef DEBUG_RF
      Serial.println( "Packet not processed & ignored." );
#endif
      rcv_Packet_Stat = -1;
      break;

    case 10:  // auto poll Mifare 1K
      read_val = Serial1.read();
      rcv_Mifare1K_ID |= read_val;
      data_to_read--;
      if ( data_to_read <= 0 ) {
        rcv_Packet_Stat = 11; // finished IDing
        break;
      }
      rcv_Mifare1K_ID <<= 8;
      break;
    case 11:
#ifdef DEBUG_RF
      Serial.println( "Mifare 1K auto poll not processed & ignored." );
#endif
      rcv_Packet_Stat = -1;
      break;
  }
}

inline boolean packet_finished() {
  return (rcv_Packet_Stat == 8);
}

inline boolean mf1k_id() {
  return (rcv_Packet_Stat == 11);
}

inline void packet_finished_clear() {
  rcv_Packet_Stat = -1;
}
// MF1K end

// Pressure sensor 4 x piezo
inline void process_pressure() {
  // read the analog in value pressure value:
  if ( millis() - fps_ts >= fps_interval ) {
    if ( !packet_is_sending() ) { // no packet is sending
      packet_sending(); // lock packet

      fps_ts = millis();

      // Clear it for recalculation
      sensor_sum = 0;

      for ( int i = 0; i < sensorNum; i++ ) {
        analogVal[i] = analogRead(A_pins[i]);
        sensor_sum += analogVal[i];
      }

      OSCMessage msg("/P");
      for (int i = 0; i < sensorNum; i++) {
        msg.add((int32_t)analogVal[i]);
      }

      SLIPSerial.beginPacket(); // mark the beginning of the OSC Packet
      msg.send(SLIPSerial);
      SLIPSerial.endPacket();
    }
  }
}
// Pressure sensor ends

// Ir distance sensor @ 100fps
inline void process_dist() {
  if ( millis() - ir_dist_ts >= ir_dist_interval ) {
    if ( !packet_is_sending() ) { // no packet is sending
      packet_sending(); // lock packet
      ir_dist_ts = millis();

      // Read analog into message IRVal
      OSCMessage msg("/I");
      for ( int i = 0; i < 2; i++ ) {
        msg.add( (int32_t)analogRead(Dist_pins[i]) );
      }

      SLIPSerial.beginPacket(); // mark the beginning of the OSC Packet
      msg.send(SLIPSerial);
      SLIPSerial.endPacket();
    }
  }
}
// Ir ends

// Blow
//inline void process_blow() {
//  int sensor_Val = analogRead(blowPin);
//
//  // Sence whenever there's cpu time
//  if ( !has_blow ) {
//    if ( sensor_Val > 30 ) {
//      has_blow = true;
//      // ts_startblow = millis();  // Started blow
//    }
//  } else {
//    if ( sensor_Val <= 15 ) {
//      has_blow = false;
//    }
//  }
//
//  if ( has_blow ) { // count up
//    samples_got++;
//
//    // count samples within 200(ms) | blow_length
//    if ( millis() - ts_startblow >= blow_length ) {
//
//      if ( samples_got > sample_trg ) { // send packet
//        if ( !packet_is_sending() ) { // no packet is sending
//          // clear sample for next perioud
//          // Needs to be inside else would miss the send
//          samples_got = 0;
//          ts_startblow = millis();
//
//          packet_sending(); // lock packet
//
//          OSCMessage msg("/Blow");
//          msg.add((int32_t)1);
//
//          SLIPSerial.beginPacket(); // mark the beginning of the OSC Packet
//          msg.send(SLIPSerial);
//          SLIPSerial.endPacket();
//        }
//      } else {
//        // Also clears when not enough sample
//        samples_got = 0;
//        ts_startblow = millis();
//      }
//    }
//  }
//}
// Blow end

inline boolean packet_is_sending() {
  if ( packet_otw ) {
    if ( millis() - packet_ts >= packet_interval ) {
      packet_ts = millis();
      packet_otw = false; // not busy any more
    }
  }
  return packet_otw;
}

inline void packet_sending() {
  packet_otw = true;
  packet_ts = millis();  // serial busy
}

inline void process_neopixel() {
  boolean start_pixel = false;
  boolean update_frame = false;

  if ( sensor_sum > 500 && neo_ID) {
    neo_ID = false;
    start_pixel = true;
  }

  // Dimm
  if ( millis() - neo_dimm_ts >= neo_dimm_interval ) {
    neo_dimm_ts = millis();
    update_frame = true;

    for ( int i = 0; i < NUMPIXELS; i++ ) {
      uint32_t col_32 = pixels.getPixelColor(i);
      if ( col_32 == 0 ) {
        continue; // save the calculation time
      }

      uint8_t rgb[3];
      for ( int j = 0; j < 3; j++ ) {
        rgb[2 - j] = col_32 & 0x000000ff;
        col_32 >>= 8;

        rgb[2 - j] = (uint8_t)(decay_ratio * ((int16_t)(rgb[2 - j])));
        if ( rgb[2 - j] < decay_thresh ) {
          rgb[2 - j] = 0;
        }
      }

      pixels.setPixelColor(i, pixels.Color(rgb[0], rgb[1], rgb[2]));
    }
  }

  switch (neo_stat) {
    case -1:  // idle
      if ( start_pixel ) {
        neo_pos = 0;  // start from centre

        neo_ts = millis() - neo_interval;
        neo_stat = 0;
      }
      break;
    case 0:
      if ( millis() - neo_ts >= neo_interval) {
        neo_ts = millis();
        pixels.setPixelColor( 3 - neo_pos, pixels.Color( 255, 255, 0 ) );
        pixels.setPixelColor( 4 + neo_pos, pixels.Color( 0, 255, 255 ) );
        update_frame = true;
        neo_pos++;
        if (neo_pos >= 4) {
          neo_stat = -1; // idle again
        }
      }
      break;
  }
  if (update_frame) {
    pixels.show();
  }
}

